/*****************************************************************************
    logfun.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

#include "params.h"
#include "globals.h"
#include "util.h"
#include "logfun.h"

#ifdef HAVE_VSNPRINTF
#define MSGBUFSIZ	256
#else
#define MSGBUFSIZ	8192
#endif

#ifndef SYSLOG_NAMES
typedef struct _code {
	char	*c_name;
	int	c_val;
} CODE;

CODE prioritynames[] =
  {
    { "alert", LOG_ALERT },
    { "crit", LOG_CRIT },
    { "debug", LOG_DEBUG },
    { "emerg", LOG_EMERG },
    { "err", LOG_ERR },
    { "error", LOG_ERR },		/* DEPRECATED */
    { "info", LOG_INFO },
    { "notice", LOG_NOTICE },
    { "panic", LOG_EMERG },		/* DEPRECATED */
    { "warn", LOG_WARNING },		/* DEPRECATED */
    { "warning", LOG_WARNING }
  };

CODE facilitynames[] =
  {
    { "auth", LOG_AUTH },
#ifdef LOG_AUTHPRIV
    { "authpriv", LOG_AUTHPRIV },
#endif
    { "cron", LOG_CRON },
    { "daemon", LOG_DAEMON },
    { "kern", LOG_KERN },
    { "lpr", LOG_LPR },
    { "mail", LOG_MAIL },
    { "news", LOG_NEWS },
    { "security", LOG_AUTH },		/* DEPRECATED */
    { "syslog", LOG_SYSLOG },
    { "user", LOG_USER },
    { "uucp", LOG_UUCP },
    { "local0", LOG_LOCAL0 },
    { "local1", LOG_LOCAL1 },
    { "local2", LOG_LOCAL2 },
    { "local3", LOG_LOCAL3 },
    { "local4", LOG_LOCAL4 },
    { "local5", LOG_LOCAL5 },
    { "local6", LOG_LOCAL6 },
    { "local7", LOG_LOCAL7 }
  };
#endif

struct lvl2pri_s { 	int level;
			int priority;	};

struct lvl2pri_s lvl2pri[] = {	{ LVL_ERR,	LOG_ERR     },
				{ LVL_WARNING,	LOG_WARNING },
				{ LVL_INFO,	LOG_NOTICE  },
				{ LVL_NOTICE,	LOG_INFO    },
				{ LVL_DEBUG,	LOG_DEBUG   }	};

#define maxlev	(int)(sizeof(lvl2pri)/sizeof(lvl2pri[0]))

int loglvl2pri(int level)
{
    if (level < 1)
        level = 1;

    if (level > maxlev)
	level = maxlev;

    return lvl2pri[level-1].priority;
}
   

int logpri2lvl (int priority)
{
    int i;

    for (i=0; i<maxlev; i++)
        if (priority == lvl2pri[i].priority)
            return lvl2pri[i].level;

    return LVL_ERR;
}


static char *getPriorityByValue(int val)
{
    int i;

    for (i = 0; i < (int)(sizeof(prioritynames)/sizeof(prioritynames[0])); i++)
        if (prioritynames[i].c_val == val)
            return (prioritynames[i].c_name);

    return "unknown";
}

int getFacilityByName(char *name)
{
    int i;

    if (name == NULL)
        return -1;

    for (i = 0; i < (int)(sizeof(facilitynames)/sizeof(facilitynames[0])); i++)
        if (!stricmp(facilitynames[i].c_name, name))
            return facilitynames[i].c_val;

    return -1;
}

char *getFacilityByValue(int facility)
{
    int i;

    for (i = 0; i < (int)(sizeof(facilitynames)/sizeof(facilitynames[0])); i++)
        if (facilitynames[i].c_val == facility)
            return facilitynames[i].c_name;

    return "unknown";
}

#ifndef HAVE_MTAG
static char *replaceMtag (char *fmt, int err, int *dyn)
{
    char *p, *s, *n, *newfmt, *errtxt;
    int  c, errlen, fmtlen;

    *dyn = 0;

    for (p=fmt, c=0; *p != '\0'; p++) {
        if (*p == '%' && *(p+1) == 'm') {
            c++;
            p++;
        }
    }

    if (!c)
        return (fmt);

    errtxt = strerror(err);
    errlen = strlen(errtxt);
    fmtlen = strlen(fmt);

    newfmt = n = (char *)Malloc(fmtlen + (c * (errlen - 2)) + 1);

    if (newfmt == NULL) 
        return fmt;

    *dyn = 1;

    for (p=s=fmt; *p != '\0'; p++) {
        if (*p == '%' && *(p+1) == 'm') {
            if (p-s) {
                memcpy (n, s, p-s);
                n += (p-s);
            }
            memcpy (n, errtxt, errlen);
            n += errlen;
            s = ++p + 1;
        }
    }
    *n = '\0';
    return newfmt;
}
#endif /* HAVE_MTAG */


void err_NoDynmem(const char *func, int line)
{
#ifdef HAVE_BACKTRACE
#define BTSIZE 64
    static void *array[BTSIZE];
    int aentries;
#endif

    LOG_MSG1 (LOG_ERR, _("%s@%d: No dynamic memory"), func, line);

#ifdef HAVE_BACKTRACE
    if (logFD >= 0) {
        aentries = backtrace (array, BTSIZE);
        backtrace_symbols_fd (array, aentries, logFD);
    }
#endif
}


void stopLog(void)
{
    if (logSD >= 0) 
        closelog();

    if (logFD >= 0)
        close (logFD);

    logSD = logFD = -1;
}


void doLog (int pri, char *fmt, ...)
{
    static char msgBuf[MSGBUFSIZ], cmdtime[32], cmtime[16];
    char mark;
    time_t now;
    va_list ap;
    int ioc=0;
#ifndef HAVE_MTAG
    int dyn = 0;
#endif /* HAVE_MTAG */
   
    if (pri & LOG_MASTER) {
        if (getpid() == parentPid)
            return;
        pri &= ~(LOG_MASTER);
    } 

#ifndef HAVE_MTAG
    if ((fmt = replaceMtag (fmt, errno, &dyn)) == NULL)
        return;
#endif /* HAVE_MTAG */

    (void) time (&now);

    if (spMode) {
        va_start (ap, fmt);
        if (logPrefix) {
            mark = (msgLevel <= maxSyslog) ? '-' : ' ';
            mark = (logFile != NULL) ? ((mark == ' ') ? '|' : '+') : mark;
            strftime (cmtime, sizeof(cmtime), "%H:%M:%S", localtime(&now));
            ioc = snprintf (msgBuf, MSGBUFSIZ, "%c%8s[%d]%c: ",
                            mark, cmtime, 
                            getpid(), *getPriorityByValue(pri));
        }
#ifdef HAVE_VSNPRINTF
        ioc += vsnprintf (&msgBuf[ioc], MSGBUFSIZ-ioc, fmt, ap);
#else
        ioc += vsprintf (&msgBuf[ioc], fmt, ap);
#endif
        msgBuf[ioc++] = '\n';

        fwrite (msgBuf, 1, ioc, stderr);
        fflush (stderr);
        va_end (ap);
    }

    if ((logSD != -1) && (logpri2lvl(pri) <= maxSyslog)) {
        va_start (ap, fmt);
#ifdef HAVE_VSNPRINTF
        ioc = vsnprintf (msgBuf, MSGBUFSIZ-ioc, fmt, ap);
#else
        ioc = vsprintf (msgBuf, fmt, ap);
#endif
        syslog(pri, msgBuf);
        va_end (ap);
    }

    if (logFD != -1) {
        va_start (ap, fmt);
        strftime (cmdtime, sizeof(cmdtime), "%b %d %H:%M:%S", localtime(&now));
        ioc = snprintf (msgBuf, MSGBUFSIZ, "%15s [%d] %-7s: ",
                        (cmdtime == NULL) ? "" : cmdtime,
                        getpid(), getPriorityByValue(pri));

#ifdef HAVE_VSNPRINTF
        ioc += vsnprintf (&msgBuf[ioc], MSGBUFSIZ-ioc, fmt, ap);
#else
        ioc += vsprintf (&msgBuf[ioc], fmt, ap);
#endif
        msgBuf[ioc++] = '\n';
        write (logFD, msgBuf, ioc);
        va_end (ap);
    }

#ifndef HAVE_MTAG
        if (dyn)
            Free(fmt);
#endif /* HAVE_MTAG */

}


void startLog(int mode)
{
    int Ferr=0;
    char *lf = logFile, *txlf = NULL;

    if ((lf != NULL) && (*lf == '!')) 
        lf++;

    if (mode & INIT_LOGGER) {

        stopLog();

        if (logFile != NULL) {
            int mode;
            if (*logFile == '!') 
                mode = O_WRONLY|O_TRUNC|O_NDELAY;
            else
                mode = O_WRONLY|O_APPEND|O_NDELAY;
            if ((txlf = xlatePath(lf)) != NULL) {
                if ((logFD = open(txlf, mode)) == -1) {
                    logFD = open(txlf, O_WRONLY|O_CREAT|O_NDELAY, 0664);
                    if (logFD == -1)
                        Ferr = 1;
                }
                else {
                    if (chmod(txlf, 0664))
                        Ferr = 1;
                }
            }
            else {
                Ferr = 1;
            }
        }
        else {
            logFD = -1;
        }

        if (maxSyslog) {
            openlog(myName, LOG_PID|LOG_NDELAY, getFacilityByName(LOGFACILITY));
            logSD = 100;
        }

        if (Ferr && (logSD != -1))
            doLog (LOG_ERR, "Cannot open Logfile '%s': %m", txlf);

        if ((logFD != -1) && (getpid() == parentPid) && (msgLevel >= LVL_INFO))
            doLog (LOG_INFO, "Opened Logfile '%s' level %d", txlf, msgLevel);
    }

    if ((mode & UPDATE_LOGGER) && (logFD != -1)) {
        if (chown(txlf, childUid, childGid))
            doLog (LOG_ERR, "Cannot set logfile '%s' ownership: %m", txlf);
        closelog();
        if (maxSyslog) {
            openlog(myName, LOG_PID|LOG_NDELAY, logFac);
            logSD = 100;
        }

    }
    if (txlf != NULL)
        Free (txlf);
}

#ifdef TEST
#include "readconfig.h"

int main (int argc, char **argv)
{
    char prio[10]; char message[200];

    evalCommandLine(argc, argv, 1);

    startLog();
    
    while (1) {
        printf("Enter priority>");
        if (fgets(prio, sizeof(prio), stdin) == NULL) break;
        printf("Enter message >");
        if (fgets(message, sizeof(message), stdin) == NULL) break;
        doLog(atoi(prio), "%s", message);
    }
    return 0;
}

#endif
